/* $XConsortium: get_load.c /main/37 1996/03/09 09:38:04 kaleb $ */
/*

Copyright (c) 1989  X Consortium

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall
not be used in advertising or otherwise to promote the sale, use or
other dealings in this Software without prior written authorization
from the X Consortium.

*/

/*
 * get_load - get system load
 *
 * Authors:  Many and varied...
 *
 * Call InitLoadPoint() to initialize.
 * GetLoadPoint() is a callback for the StripChart widget.
 */

#include <X11/Xos.h>
#include <X11/Intrinsic.h>
#include <stdio.h>

#include <nlist.h>
extern void nlist();

#include <sys/param.h>

#include <sys/utsname.h>


static void
xload_error(char *str1, char *str2)
{
  fprintf(stderr, "xload: %s %s\n", str1, str2);
  exit(-1);
}

static enum {SunOS, Linux} ostype;


#define KMEM_FILE "/dev/kmem"
#define KERNEL_FILE "/vmunix"
#define KERNEL_LOAD_VARIABLE "_avenrun"

static struct nlist namelist[] = {  /* namelist for vmunix grubbing */
#define LOADAV 0
    {KERNEL_LOAD_VARIABLE},
    {0}
};

static int kmem;
static long loadavg_seek;


inline void
InitLoadPoint_sunos()
{
  nlist(KERNEL_FILE, namelist);

  if (
    namelist[LOADAV].n_type == 0 ||
    namelist[LOADAV].n_value == 0
  ) {
    xload_error("cannot get name list from", KERNEL_FILE);
    exit(-1);
  }

  loadavg_seek = namelist[LOADAV].n_value;
  
  kmem = open(KMEM_FILE, O_RDONLY);
  if (kmem < 0) {
    xload_error("cannot open", KMEM_FILE);
  }
}


inline void
InitLoadPoint_linux()
{
}

inline void
GetLoadPoint_sunos(
  Widget   w,		/* unused */
  caddr_t  closure,	/* unused */
  caddr_t  call_data	/* pointer to (double) return value */
)
{
  double *loadavg = (double *)call_data;

  lseek(kmem, loadavg_seek, 0);

  {
    long temp;
    read(kmem, (char *)&temp, sizeof(long));
    *loadavg = (double)temp/FSCALE;
  }
}


inline void
GetLoadPoint_linux(
   Widget   w,              /* unused */
   caddr_t  closure,        /* unused */
   caddr_t  call_data       /* pointer to (double) return value */
)
{
  static int fd = -1;
  int n;
  char buf[10];

  if (fd < 0) {
    if (fd == -2 || (fd = open("/proc/loadavg", O_RDONLY)) < 0) {
      fd = -2;
      *(double *)call_data = 0.0;
      return;
    }
  } else {
    lseek(fd, 0, 0);
  }

  if (
    (n = read(fd, buf, sizeof(buf)-1)) > 0 &&
    sscanf(buf, "%lf", (double *)call_data) == 1
  ) {
    return;
  }

  *(double *)call_data = 0.0;     /* temporary hiccup */
}


void
InitLoadPoint()
{
  struct utsname un;

  uname(&un);

  if (!strcmp(un.sysname, "SunOS")) {
    ostype = SunOS;
  } else if (!strcmp(un.sysname, "Linux")) {
    ostype = Linux;
  } else {
    xload_error("unknown OS type", un.sysname);
    exit(-1);
  }

  switch (ostype) {
    case SunOS:
      InitLoadPoint_sunos();
      break;

    case Linux:
      InitLoadPoint_linux();
      break;
  }
}

void
GetLoadPoint(
   Widget   w,              /* unused */
   caddr_t  closure,        /* unused */
   caddr_t  call_data       /* pointer to (double) return value */
)
{
  switch (ostype) {
    case SunOS:
      GetLoadPoint_sunos(w, closure, call_data);
      break;

    case Linux:
      GetLoadPoint_linux(w, closure, call_data);
      break;
  }
}
